/* @execvpe.c - exec a program, searching PATH environment variable
 */
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "gen.h"

#define MAXARGV 512

#ifndef _NO_PROTOTYPE
void getshell(char *shell, char *av1, char *path);
int whencep(char *buf, char *fname);
#endif
extern char **environ;

int execvpe(char *fname, char *argv[], char **envp)
	{
	char ename[PATHSIZE+1];

#	ifdef DEB
	errfmt("execvpe: before execv fname '%s' *argv '%s'\n", fname, *argv);
#	endif
	if (whencep(ename, fname) < 0)
		;
	else
		{
		execve(ename, argv, envp ? envp : environ);
#		ifdef DEB
		errfmt("Cannot exec: errno  %d ENO %d\n", errno, ENOEXEC);
#		endif
		if (errno == ENOEXEC)
			{
			char shell[PATHSIZE+1], av1[PATHSIZE+1];
			char *argvbuf[MAXARGV+1];
			char **avptr;
			int count = 0;
			getshell(shell, av1, ename);
			argvbuf[count++] = shell;
			if (av1[0] != EOS)
				argvbuf[count++] = av1;
			argvbuf[count++] = ename;
			for (avptr = &argv[1]; *avptr != NULL && count < MAXARGV; count++, avptr++)
				argvbuf[count] = *avptr;
			if (MAXARGV <= count)
				errno = E2BIG;
			else
				{
				argvbuf[count] = NULL;
#				ifdef DEB
				errfmt("execvpe: shell %s i %d argvbuf", shell, count);
				for (avptr = argvbuf; *avptr; ++avptr)
					errfmt(" %s", *avptr);
				errfmt("\n");
#				endif
				execve(shell, argvbuf, envp ? envp : environ);
				}
			}
		}
#	ifdef DEB
	errfmt("execvpe: errno %d\n", errno);
#	endif
	return -ERRNO;
	}

/* @whencep - search for a program using PATH environment variable
 */
int whencep(char *buf, char *fname)
	{
	int rtn = -ENOENT;
	char *path, *next, *p;
	struct stat sstat;

	path = getenv("PATH");
	next = (fname[scnstr(fname, '/')] == EOS) ? path : ":";
#	ifdef DEB
	errfmt("whencep: next '%s'\n", next);
#	endif
	while (*next != EOS)
		{
		int i;

		for (p = next, i = 0; *p != EOS && *p != ':'; ++p)
			{
			if (i < PATHSIZE)
				buf[i++] = *p;
			}
		next = (*p == EOS || *(p + 1) == EOS && p != next) ? p : p + 1;
		if (i < PATHSIZE && i != 0)
			buf[i++] = '/';
		for (p = fname; i < PATHSIZE && (buf[i] = *p++) != EOS; ++i)
			;
#		ifdef DEB
		errfmt("whencep: before access buf '%s'\n", buf);
#		endif
		if (0 <= access(buf, F_OK))
			if (0 <= access(buf, X_OK) && 0 <= stat(buf, &sstat) && S_ISREG(sstat.st_mode))
				{
				rtn = 0;
				break;
				}
			else
				rtn = -EACCES;
		}
	errno = -rtn;
#	ifdef DEB
	errfmt("whencep: rtn %d errno %d\n", rtn, errno);
#	endif
	return rtn;
	}

/* @getshell - work out which shell to use interpreting #! 
			   for operating systems that do not support it.
 */
void getshell(char *shell, char *av1, char *path)
	{
	static char *defshell = "/bin/sh";
	char buf[32+1], *ptr;
	int res;
	FD fd;
	Bool ok = NO;

#	ifdef DEB
	errfmt("getshell:\n");
#	endif
	if (0 <= (fd = open(path, O_RDONLY, 0)))
		{
		buf[(0 <= (res=read(fd, buf, sizeof(buf)-1))) ? res : 0] = EOS;
		if (buf[0] == '#' && buf[1] == '!')
			{
			for (ptr = &buf[2]; isspace(*ptr); ptr++)
				;
			while (*ptr != EOS && !isspace(*ptr))
				*shell++ = *ptr++;
			*shell = EOS;
			while (*ptr != EOS && *ptr != '\n' && isspace(*ptr))
				ptr++;
			if (*ptr != EOS && *ptr != '\n')
				while (!isspace(*ptr))
					*av1++ = *ptr++;
			*av1 = EOS;
			ok = YES;
			}
		close(fd);
		}
	if (!ok)
		{
		if ((ptr = getenv("SHELL")) != NULL)
			strcpy(shell, ptr);
		else
			strcpy(shell, defshell);
		}
#	ifdef DEB
	errfmt("getshell:1 '%s' av1 '%s'\n", shell, av1);
#	endif
	}

