/* @gen.c - general functions
 */
#define VAR_SRC
#include "gen.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <setjmp.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "bopen.h"

char *progname = "";

/* b@get - read a line from a file
 */
int get(BFD bfd, char *buf, int size)
	{
	int count = 0, res = 1;

	while (count < size - 1)
		{
		if ((res = bread(bfd, &buf[count], 1)) < 0)
			return -ERRNO;
		else if (res == 0)
			break;
		if (buf[count] == '\n')
			{
			count++;
			break;
			}
#		ifdef DEB
		errfmt("get:read '%c'\n", buf[count]);
#		endif
		count++;
		}
	buf[count] = EOS;
#	ifdef DEB
	errfmt("get: count %d '%s'\n", count, buf);
#	endif
	return count;
	}

/* b@cgetfd - get one character from buffered file
 */
int cgetfd(BFD bfd)
	{
	int rtn;
	char c;

	if (bread(bfd, &c, 1) <= 0)
		rtn = EOF;
	else
		rtn = (int)c;
	return rtn;
	}

/* b@cputfd - put one character to buffered file
 */
int cputfd(BFD bfd, int ichar)
	{
	int rtn;
	char c;

	c = (char)ichar;
	if (bwrite(bfd, &c, 1) <= 0)
		ichar = EOF;
	return ichar;
	}

/* b@vfprint - formatted output to buffered file
 */
int vfprint(BFD bfd, char *fmt, va_list list)
	{
	int rtn = 0;
	char buf[BUFSIZ];

#	ifdef DEB
	errfmt("vfprint: '%s'\n", fmt);
#	endif
# 	if defined(sparc)
#	define SUNHACK1 (int)
#	define SUNHACK2 rtn = strlen(buf);
#	else
#	define SUNHACK1
#	define SUNHACK2
#	endif

	rtn = SUNHACK1 vsprintf(buf, fmt, list); SUNHACK2
	bwrite(bfd, buf, strlen(buf));
#	ifdef DEB
	errfmt("vfprint:end '%s'\n", fmt);
#	endif
	return rtn;
	}

/* g@t_basename - last component in path
 */
char *t_basename(char *buf, char *path)
	{
	char *ptr, *orig;

	orig = buf;
	for (ptr = path; *ptr; ptr++)
		;
	for ( ; ptr != path && *ptr != '/'; ptr--)
		;
	if (*ptr == '/')
		ptr++;
	while (*ptr)
		*buf++ = *ptr++;
	*buf = '\0';
	return orig;
	}

/* g@rmsuff - remove suffix from path
 */
char *rmsuff(char *buf, char *suff)
	{
	int lensuff, lenbuf;

	lensuff = strlen(suff);
	lenbuf = strlen(buf);
	if (lensuff <= lenbuf)
		{
		if (!strncmp(&buf[lenbuf - lensuff], suff, lensuff))
			{
			buf[lenbuf - lensuff] = EOS;
			return buf;
			}
		}
	return NULL;
	}

/* g@dirname - directory component of path
 */
char *dirname(char *buf, char *path)
	{
	char *ptr, *orig;

	orig = buf;
	for (ptr = path; *ptr; ptr++)
		;
	for ( ; ptr != path && *ptr != '/'; ptr--)
		;
	if (ptr == path && *ptr == '/')
		ptr++;
	while (path != ptr)
		*buf++ = *path++;
	if (buf == orig)
		{
		*buf++ = '.';
		*buf = '\0';
		}
	else
		*buf = '\0';
	return orig;
	}

/* g@fmalloc - debug malloc
 */
char *fmalloc(int size)
	{
#	ifdef DEB
	char *rtn;
	errfmt("fmalloc:0 size %d\n", size);
	rtn = (char *)malloc(size);
	errfmt("fmalloc: size %d rtn %d\n", size, rtn);
	return rtn;
#	else
	return (char *)malloc(size);
#	endif
	}

/* g@normpath - remove ".."s and "."s from a path name
 */
char *normpath(char *path)
	{
	Bool root = NO;
	char *ptr, *cptr;
	int diff;

	if (*path == '/')
		root = YES;
	ptr = path;
	while (*ptr)
		{
#		ifdef DEB
		errfmt("1:%s ptr '%s'\n", path, ptr);
#		endif
		if (!strncmp(ptr, "/../", 4) || !strcmp(ptr, "/.."))
			{
			cptr = ptr;
			if (*cptr == '/')
				while (path < cptr && *cptr == '/')
					cptr--;
			if (path < cptr)
				for (cptr--; path < cptr && *cptr != '/'; )
					cptr--;
			ptr += 3;
			diff = cptr - ptr;
#			ifdef DEB
			errfmt("diff:%d ptr '%s' -5 '%s' \n", diff, ptr, ptr + diff);
#			endif
			while (*ptr)
				{
#				ifdef DEB
				errfmt("diff.1:%d ptr '%s' -5 '%s' \n", diff, ptr, ptr + diff);
#				endif
				ptr[diff] = *ptr;
				ptr++;
				}
#			ifdef DEB
			errfmt("diff0:%d ptr '%s' -5 '%s' '%c' '%c'\n", diff, ptr, ptr + diff, *ptr, ptr[diff]);
#			endif
			ptr[diff] = '\0';
#			ifdef DEB
			errfmt("diff1:%d ptr '%s' -5 '%s' '%c' '%c'\n", diff, ptr, ptr + diff, *ptr, ptr[diff]);
#			endif
			ptr = path;
			}
		else if (!strncmp(ptr, "/./", 3) || !strcmp(ptr, "/."))
			{
			for (ptr += 2; *ptr; ptr++)
				ptr[-2] = *ptr;
			ptr[-2] = *ptr;
			ptr = path;
			}
		else
			{
			ptr++;
			while(*ptr && *ptr != '/')
				ptr++;
			}
		}
	if (!*path && root)
		strcpy(path, "/");
	return path;
	}

/* g@cpyfld - extract field in string
 */
char *cpyfld(char *buf, short fnumber, char *field, char delim)
	{
	short i;
	char *p, *q;

	p = buf;
	q = field;
	for(i = 1; i < fnumber && *p != '\n' && *p != EOS; ++i)
		{
		while(*p != delim && *p != '\n' && *p != EOS)
			++p;
		++p;
		}
	if (*p == '\n' || *p == EOS)
		{
		return (NULL);
		}
	else
		{
		while(*p != delim && *p != '\n' && *p != EOS)
			*q++ = *p++;
		*q = EOS;
		if (*p == delim)
			return (++p);
		else
			return (p);
		}
	}

/* @nextpos - find next component of path
 */
int nextpos(char *path, int pos)
	{
#	ifdef DEB
	errfmt("nextpos: path '%s' pos %d\n", path, pos);
#	endif
	for (; path[pos] == '/'; pos++)
		;
	if (path[pos])
		for (; path[pos] != '/' && path[pos] != EOS; pos++)
			;
	else
		pos = -1;
	return pos;
	}

/* @getfd - get a new file descriptor
 */
FD getfd(void **ptr, int maxfd)
	{
	int count = 0;
	for (; count < maxfd; ptr++, count++)
		if (!*ptr)
			return count;
	return -(errno = EMFILE);
	}

/* @getobj - validate a file descriptor
 */
void *getobj(void **ptr, int maxfd, int fd)
	{
	void *rtn;

	if (fd < 0 || maxfd <= fd)
		rtn = NULL;
	else
		rtn = ptr[fd];
	if (rtn == NULL)
		errno = EBADF;
	return rtn;
	}

/* g@lower - convert a string to lower case
 */
void lower(char *str)
	{
	while (*str)
		{
		*str = tolower(*str);
		str++;
		}
	}

/* g@upper - convert a string to upper case
 */
void upper(char *str)
	{
	while (*str)
		{
		*str = toupper(*str);
		str++;
		}
	}

/* g@rsetjmp - version of setjmp that allows returning 0
 */
int rsetjmp(jmp_buf env)
	{
	int rtn;
	if ((rtn = setjmp(env)) == RJMPINIT)
		rtn = JMPINIT;
	else if (rtn == JMPINIT)
		rtn = RJMPINIT;
	return rtn;
	}

/* g@rlongjmp - version of longjmp that allows returning 0
 */
void rlongjmp(jmp_buf env, int rtn)
	{
	if (rtn == JMPINIT)
		rtn = RJMPINIT;
	longjmp(env, rtn);
	}

/* g@errmesg - return a pointer to an error message string
 */
char *errmesg(int errno)
	{
#if defined(_POSIX_SOURCE)
	static char buf[20];
	sprintf(buf, "OS Errno %d", errno);
	return buf;
#else
	return sys_errlist[errno];
#endif
	}

char *strtooct(char *buf, char *str, Bool trybackslash)
	{
	char c;
	while (*str)
		{
		c = *str;
		if (c == '\n' || (' ' <= c  && c < 0177))
			*buf++ = c;
		else if (trybackslash && c < 14)
			{
			*buf++ = '\\';
			*buf++ = "01234567btnvfr"[c];
			}
		else
			{
			*buf++ = '\\';
			*buf++ = (c >> 6) & 07 | '0';	/*	ASCII ***/
			*buf++ = (c >> 3) & 07 | '0';
			*buf++ = c & 07 | '0';
			}
		str++;
		}
	*buf = EOS;
	return buf;
	}

char *strtoquest(char *buf, char *str)
	{
	while (*str)
		{
		if (*str == '\n' || ' ' <= *str && *str < 0177)
			*buf++ = *str;
		else
			*buf++ = '?';
		str++;
		}
	*buf = EOS;
	return buf;
	}

int printifd(BFD bfd, long l, int len, int base)
	{
	char buf[BUFSIZ+1];
	int res, count, num;

	buf[res = ltob(buf, l, base)] = EOS;
	count = res;
	for (; res < len; res++)
		cputfd(bfd, ' ');
	num = res - count;
	for (res = 0; res < count; res++)
		cputfd(bfd, buf[res]);
	return num + res;
	}

int printbfd(BFD bfd, char *str, int len)
	{
	int res;
	bwrite(bfd, str, res = strlen(str));
	for (; res < len; res++)
		cputfd(bfd, ' ');
	return res;
	}

/* ixz@_testperm - calculate whether permission is available
 */
int _testperm(Mode mode, uid_t uid, gid_t gid, int omode, uid_t myuid, gid_t mygid)
	{
	int rtn = 0;
	
	if (myuid != ROOTUID)
		{
		if (myuid == uid)
			{
			if (!(mode & S_IRUSR) && omode == R_OK)
				rtn = -EACCES;
			if (!(mode & S_IWUSR) && omode == W_OK)
				rtn = -EACCES;
			if (!(mode & S_IXUSR) && omode == X_OK)
				rtn = -EACCES;
			}
		else if (mygid == gid)
			{
			if (!(mode & S_IRGRP) && omode == R_OK)
				rtn = -EACCES;
			if (!(mode & S_IWGRP) && omode == W_OK)
				rtn = -EACCES;
			if (!(mode & S_IXGRP) && omode == X_OK)
				rtn = -EACCES;
			}
		else
			{
			if (!(mode & S_IROTH) && omode == R_OK)
				rtn = -EACCES;
			if (!(mode & S_IWOTH) && omode == W_OK)
				rtn = -EACCES;
			if (!(mode & S_IXOTH) && omode == X_OK)
				rtn = -EACCES;
			}
		}
#	ifdef DEB
	errfmt("testperm: mode %d uid %d gid %d omode %d rtn %d myuid %d mygid %d\n", mode, uid, gid, omode, rtn, myuid, mygid);
#	endif
	if (rtn == -EACCES)
		errno = EACCES;
	return rtn;
	}

static Bool errprint = NO;
static Bool errshowpid = NO;

/* g@errfmt - print debug
 */
int errfmt(char *fmt, VAR_ARGS)
	{
	int rtn = 0;
	va_list list;
	static Bool tested = NO;

	if (!tested)
		{
		errprint = (getenv("ERRFMT") && *getenv("ERRFMT"));
		errshowpid = (getenv("ERRSHOWPID") && *getenv("ERRSHOWPID"));
		tested = YES;
		}
	if (errprint)
		{
		char buf[BUFSIZ];
		va_start(list, fmt);
		if (errshowpid)
			{
			sprintf(buf, "%d:", getpid());
			write(STDERR_FILENO, buf, strlen(buf));
			}
		rtn = SUNHACK1 vsprintf(buf, fmt, list); SUNHACK2
		write(STDERR_FILENO, buf, strlen(buf));
		va_end(list);
		}
	return rtn;
	}
void errexit(int cerrno, char *fmt, VAR_ARGS)
	{
	va_list list;
	va_start(list, fmt);
	vfprint(STDERR_FILENO, fmt, list);
	va_end(list);
	errno = cerrno;
	exit(errno);
	}
int failure(int cerrno, char *fmt, VAR_ARGS)
	{
	va_list list;
	va_start(list, fmt);
	vfprint(STDERR_FILENO, fmt, list);
	va_end(list);
	if (cerrno)
		{
		errno = cerrno;
		perror(progname);
		}
	return -cerrno;
	}

