/* @genstd.c - general functions that depend on whether virtual file systems are in use
 */
#include "gen.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#include "rbopen.h"
#ifndef STANDARD
#include "zstand.h"
#endif

char *pname(char *buf, char *new, char *dir)
	{
	Stat sstat;

	if (strchr(new, '/') == NULL && 0 <= stat(dir, &sstat) && S_ISDIR(sstat.st_mode))
		cpystr(buf, dir, "/", new, NULL);
	else
		strcpy(buf, dir);
#	ifdef DEB
	errfmt("pname: buf '%s' new '%s' dir '%s'\n", buf, new, dir);
#	endif
	return buf;
	}

/* g@numindir - count the number of entries in a directory
 */
int numindir(DIR *dir)
	{
	int rtn = 0;
	Seek save;
#if !defined(_NO_PROTOTYPE)
	extern Seek telldir(DIR *);
	extern struct dirent *readdir(DIR *);
#endif

#if !defined(_POSIX_SOURCE)
	save = telldir(dir);
#endif
	rewinddir(dir);
	while (readdir(dir) != NULL)
		rtn++;
#if !defined(_POSIX_SOURCE)
	seekdir(dir, save);
#else
	rewinddir(dir);
#endif
#	ifdef DEB
	errfmt("numindir: %d\n", rtn);
#	endif
	return rtn;
	}

/* z@rreadlink - readlink, null terminated
 */
int rreadlink(char *path, char *buf, int bufsiz)
	{
	int size;

	if (0 <= (size = readlink(path, buf, bufsiz)))
		buf[size] = EOS;
	else
		buf[0] = EOS;
#	ifdef DEB
	errfmt("rreadlink: '%s' size %d\n", buf, size);
#	endif
	return size;
	}

/* z@fprint - formatted output to file descriptor
 */
int fprint(BFD bfd, char *fmt, VAR_ARGS)
	{
	int rtn = 0;
	va_list list;

#	ifdef DEB
	errfmt("fprint:beg\n");
#	endif
	va_start(list, fmt);
	rtn = vfprint(bfd, fmt, list);
	va_end(list);
#	ifdef DEB
	errfmt("fprint:end\n");
#	endif
	return rtn;
	}

/* z@print - formatted output to stdout
 */
int print(char *fmt, VAR_ARGS)
	{
	int rtn = 0;
	va_list list;

#	ifdef DEB
	errfmt("print:beg\n");
#	endif
	va_start(list, fmt);
	rtn = vfprint(BSTDOUT, fmt, list);
	va_end(list);
#	ifdef DEB
	errfmt("print:end\n");
#	endif
	return rtn;
	}
		
char *passwdnamenum(char *buf, uid_t num, Bool shownum)
	{
	struct passwd *pwd;

#	ifdef DEB
	errfmt("passwdnamenum:beg %d\n", num);
#	endif
	if (!shownum && (pwd = getpwuid(num)) != NULL)
		strcpy(buf, pwd->pw_name);
	else
		sprintf(buf, "%d", num);
#	ifdef DEB
	errfmt("passwdnamenum:end %s\n", buf);
#	endif
	return buf;
	}

char *groupnamenum(char *buf, gid_t num, Bool shownum)
	{
	struct group *grp;

#	ifdef DEB
	errfmt("groupnamenum:beg %d\n", num);
#	endif
	if (!shownum && (grp = getgrgid(num)) != NULL)
		{
#	ifdef DEB
	errfmt("groupnamenum:1 %d\n", grp->gr_name);
#	endif
#	ifdef DEB
	errfmt("groupnamenum:2 '%s'\n", grp->gr_name);
#	endif
		strcpy(buf, grp->gr_name);
#	ifdef DEB
	errfmt("groupnamenum:3 '%s'\n", buf);
#	endif
		}
	else
		sprintf(buf, "%d", num);
#	ifdef DEB
	errfmt("groupnamenum:end %s\n", buf);
#	endif
	return buf;
	}

int mkdirall(char *path, Mode mode)
	{
	int rtn = 0;
	Stat sstat;
	char npath[PATHSIZE+1];

#	ifdef DEB
	errfmt("mkdirall: path '%s'\n", path);
#	endif
	if (strcmp(path, "/") != 0 && strcmp(path, ".") != 0) 
		if (stat(path, &sstat) < 0)
			{
			if (mkdirall(dirname(npath, path), mode) < 0)
				rtn = -ERRNO;
			else if (mkdir(path, mode) < 0)
				{
				rtn = -ERRNO;
				}
			}
		else
			rtn = -(errno = EEXIST);
#	ifdef DEB
	errfmt("mkdirall: path '%s' rtn %d\n", path, rtn);
#	endif
	return rtn;
	}
